module Diff3 where

import Diff
import Types

data Origin
  = Stable
  | Mine
  | Your
  deriving (Show, Eq)

--diff3Toks :: TV -> TV -> TV -> Diff
diff3Toks tMine tOrig tYour =
  conflict $ align (diffToks tOrig tMine) (diffToks tOrig tYour)
  where
    align :: Diff -> Diff -> [(Origin, (Op, Tok))]
    align ((Keep, a):as) ((Keep, _):bs) = (Stable, (Keep, a)) : align as bs
    align ((Add, a):as) ((Add, b):bs) =
      (Mine, (Add, a)) : (Your, (Add, b)) : align as bs
    align ((Remove, a):as) ((Remove, b):bs) =
      (Mine, (Remove, a)) : (Your, (Remove, b)) : align as bs
    align ((Add, a):as) bs@((Keep, _):_) = (Mine, (Add, a)) : align as bs
    align as@((Keep, _):_) ((Add, b):bs) = (Your, (Add, b)) : align as bs
    align ((Remove, a):as) ((Keep, b):bs) =
      (Mine, (Remove, a)) : (Your, (Keep, b)) : align as bs
    align ((Keep, a):as) ((Remove, b):bs) =
      (Mine, (Keep, a)) : (Your, (Remove, b)) : align as bs
    align ((Add, a):as) bs@((Remove, _):_) = (Mine, (Add, a)) : align as bs
    align as@((Remove, _):_) ((Add, b):bs) = (Your, (Add, b)) : align as bs
    align [] [] = []
    align as@((Add, _):_) [] = map ((,) Mine) as
    align [] bs@((Add, _):_) = map ((,) Your) bs
    align _ _ = error "Internal failure: diffstreams seem broken, cannot align"
    conflict :: [(Origin, (Op, Tok))] -> Diff
    conflict [] = []
    conflict as@(a:_)
      | stable a = applySplit stable (map snd) conflict as
      | unstable a = applySplit unstable merge conflict as
    applySplit :: (a -> Bool) -> ([a] -> [b]) -> ([a] -> [b]) -> [a] -> [b]
    applySplit cond f g xs = f (takeWhile cond xs) ++ g (dropWhile cond xs)
    merge :: [(Origin, (Op,Tok))] -> Diff
    merge cs =
      let mys = map (\a -> map snd $ filter ((== a) . fst) cs) [Mine, Your]
          [tokOrigsMine, tokOrigsYour] =
            map (map snd.filter ((/= Add) . fst)) mys
          [tokMine, tokYour] = map (map snd.filter ((/= Remove) . fst)) mys
       in if tokOrigsMine /= tokOrigsYour
            then error "Internal failure: merge origins differ"
            else map ((,) MineChanged) tokMine ++
                 map ((,) Original) tokOrigsMine ++
                 map ((,) YourChanged) tokYour
    stable (Stable, _) = True
    stable _ = False
    unstable = not . stable
