{-# LANGUAGE OverloadedStrings #-}
module Merge
  ( MergeOpts(..)
  , mergeOption
  , fmtMerged
  ) where

import qualified Data.ByteString as B
import qualified Data.ByteString.Builder as BB
import Data.String
import Options.Applicative
import Types

data MergeOpts =
  MergeOpts
    { mergeDoMerge :: Bool
    , mergeForceWhitespace :: Bool
    , mergeKeepWhitespace :: Bool
    , mergeCStartStr :: BS
    , mergeMineSepStr :: BS
    , mergeYourSepStr :: BS
    , mergeCEndStr :: BS
    }
  deriving (Show)

marker = fromString . replicate 7

mergeOption forPatch = addLBR <$> ((,) <$> switch (short 'a' <> long "add-linebreak" <> help "Automatically add a line break after conflict markers") <*> mo)
  where
   mo = MergeOpts <$>
    switch
      (short 'm' <>
       long "merge" <>
       help
         (if forPatch
            then "Merge using conflict markers instead of printing the rejected hunks"
            else "Output the merged file instead of the patch")) <*>
    switch
      (short 'w' <>
       long "whitespace" <>
       help
         ((if forPatch
             then "Force rejecting a thunk"
             else "Force a merge conflict") ++
          " on whitespace mismatch")) <*>
    switch
      (short 'k' <>
       long "keep-whitespace" <>
       help
         ("On whitespace mismatch, default to the version from " ++
          (if forPatch
             then "original file"
             else "MYFILE") ++
          " instead of the one from " ++
          (if forPatch
             then "patch"
             else "YOURFILE"))) <*>
    strOption
      (long "merge-start" <>
       value (marker '<') <> help "Marker for the beginning of a conflict") <*>
    strOption
      (long "merge-mine" <>
       value (marker '|') <>
       help "Marker that separates `mine' from `original' part of the conflict") <*>
    strOption
      (long "merge-your" <>
       value (marker '=') <>
       help "Marker that separates `original' from `your' part of the conflict") <*>
    strOption
      (long "merge-end" <>
       value (marker '>') <> help "Marker for the end of a conflict")
   addLBR (False,x)=x
   addLBR (True,x)=x{
    mergeCStartStr=mergeCStartStr x <> "\n",
    mergeMineSepStr=mergeMineSepStr x <> "\n",
    mergeYourSepStr=mergeYourSepStr x <> "\n",
    mergeCEndStr=mergeCEndStr x <> "\n"}

{- This kinda relies on reasonable ordering within the conflicts in the Diff -}
fmtMerged :: MergeOpts -> Diff -> BB.Builder
fmtMerged mo = go Keep
  where
    go op []
      | conflictOp op = bb $ mergeCEndStr mo
      | otherwise = mempty
    go last l@((op, (_, tok)):xs)
      | conflictOp last && not (conflictOp op) =
        bb (mergeCEndStr mo) <> go Keep l
      | not (conflictOp last) && conflictOp op =
        bb (mergeCStartStr mo) <> go MineChanged l
      | last /= op && conflictOp op =
        (case op of
           MineChanged -> bb $ mergeCStartStr mo
           Original -> bb $ mergeMineSepStr mo
           YourChanged -> bb $ mergeYourSepStr mo) <>
        go op l
      | otherwise = bb tok <> go op xs
    conflictOp o =
      case o of
        Keep -> False
        Add -> False
        Remove -> False
        _ -> True
    bb = BB.byteString
